import csv
import logging
from collections import Counter
from datetime import datetime, timedelta, timezone
from typing import Any

from sqlglot import MappingSchema, exp, parse_one
from sqlglot.optimizer.qualify import qualify
from sqlglot.optimizer.scope import build_scope, find_all_in_scope

from odp.core.types import Dialect, QueryRow, SchemaRow

logger = logging.getLogger(__name__)


def read_queries(
    query_file: str,
    since: int,
) -> list[QueryRow]:
    since_datetime = datetime.now(timezone.utc) - timedelta(days=since)
    # Read queries from a CSV file and return a list of dictionaries where each key is a column in the CSV
    with open(query_file) as f:
        csv_reader = csv.reader(f)
        header = list(map(str.upper, next(csv_reader)))

        rows = []
        for _row in csv_reader:
            row: dict[str, Any] = dict(zip(header, _row))
            row["START_TIME"] = datetime.fromisoformat(row["START_TIME"])
            query_row = QueryRow(**row)
            if since_datetime < query_row.START_TIME:
                rows.append(query_row)

        return rows


def read_info_schema_from_file(schema_file: str) -> tuple[dict, list[tuple]]:
    # Read the info schema from a CSV file and return it as both a nested dictionary and a flat list
    # Format is: catalog -> schema -> table name -> column name
    schema_rows: list[SchemaRow] = []
    with open(schema_file) as f:
        csv_reader = csv.reader(f)
        next(csv_reader)  # Skip header
        for row in csv_reader:
            catalog, schema_name, table_name, column_name = map(str.upper, row)
            schema_rows.append(
                SchemaRow(
                    TABLE_CATALOG=catalog,
                    TABLE_SCHEMA=schema_name,
                    TABLE_NAME=table_name,
                    COLUMN_NAME=column_name,
                )
            )

    return build_info_schema(schema_rows)


def build_info_schema(schema_rows: list[SchemaRow]) -> tuple[dict, list[tuple]]:
    sqlglot_mapping_schema: dict = {}
    flat_schema: list[tuple] = []
    for row in schema_rows:
        catalog, schema_name, table_name, column_name = (
            row.TABLE_CATALOG,
            row.TABLE_SCHEMA,
            row.TABLE_NAME,
            row.COLUMN_NAME,
        )

        if catalog not in sqlglot_mapping_schema:
            sqlglot_mapping_schema[catalog] = {}
        if schema_name not in sqlglot_mapping_schema[catalog]:
            sqlglot_mapping_schema[catalog][schema_name] = {}
        if table_name not in sqlglot_mapping_schema[catalog][schema_name]:
            sqlglot_mapping_schema[catalog][schema_name][table_name] = {}
        sqlglot_mapping_schema[catalog][schema_name][table_name][column_name] = "DUMMY"
        flat_schema.append((catalog, schema_name, table_name, column_name))

    return sqlglot_mapping_schema, flat_schema


def extract_columns(
    query_text: str,
    database_name: str | None,
    catalog_name: str | None,
    schema: dict,
    dialect: Dialect,
) -> list[tuple]:
    # Extract the columns from a query that map to actual columns in a table
    # Based on https://github.com/tobymao/sqlglot/blob/main/posts/ast_primer.md
    try:
        parsed = parse_one(query_text, dialect=dialect.value)
        qualified = qualify(
            parsed, schema=schema, dialect=dialect.value
        )  # Qualify (add schema) and expand * to explicit columns
        root = build_scope(qualified)
    except Exception as e:
        logger.debug("Skipping failed query: %s, %s", query_text, e)
        # todo - debug log these / write to file
        return []
    if root is None:
        return []

    # This is confusing due to naming conventions. We basically want to make sure every table is fully qualified
    # sqlglot has {catalog: {db: {table: {col: type}}}} convention
    # Snowflake has {database_name: {schema_name: {table: {col: type}}}}
    # So we do database_name (SF) -> catalog (sqlglot), schema_name (SF) -> db (sqlglot)
    for source in root.sources:
        s = root.sources[source]
        if type(s) == exp.Table:
            if "db" not in s.args or not s.args["db"]:
                s.set("db", exp.Identifier(this=catalog_name, quoted=True))
            if "catalog" not in s.args or not s.args["catalog"]:
                s.set("catalog", exp.Identifier(this=database_name, quoted=True))

    columns = []
    for column in find_all_in_scope(root.expression, exp.Column):
        if column.table not in root.sources:
            continue

        table = root.sources[column.table]
        if type(table) != exp.Table:
            continue

        columns.append(
            (
                table.catalog,
                table.db,
                table.name,
                column.this.this,
            )
        )
    return columns


def extract_tables(
    query_text: str,
    schema: dict | MappingSchema,
    dialect: Dialect,
    catalog_name: str | None = None,
    database_name: str | None = None,
) -> list[tuple]:
    # Extract the tables from a query that map to actual columns in a table
    # Based on https://github.com/tobymao/sqlglot/blob/main/posts/ast_primer.md
    try:
        parsed = parse_one(query_text, dialect=dialect.value)
        qualified = qualify(
            parsed,
            catalog=catalog_name,
            db=database_name,
            schema=schema,
            dialect=dialect.value,
            infer_schema=True,
            expand_stars=False,  # we don't care about columns here
            validate_qualify_columns=False,  # we don't care about columns here
            qualify_columns=False,  # we don't care about columns here
        )
        root = build_scope(qualified)
    except Exception as e:
        logger.debug("Skipping failed query: %s, %s", query_text, e)
        # todo - debug log these / write to file
        return []

    if root is None:
        return []

    table_exps = set()
    for scope in root.traverse():
        for _, (_, source) in scope.selected_sources.items():
            if isinstance(source, exp.Table):
                table_exps.add(source)

    # converting table expressions to a list of tuples
    tables = []
    for table_exp in table_exps:
        tables.append(
            (
                table_exp.catalog,
                table_exp.db,
                str(table_exp.this.name),
            )
        )
    return tables


def summarize_columns(columns: list[list[tuple]]) -> Counter:
    # Return a dictionary of column to counts

    # Flatten the col vals
    cols = [item for sublist in columns for item in sublist]
    return Counter(cols)


def detect_unused_columns(
    queries: list[QueryRow],
    info_schema: dict,
    info_schema_flat: list[tuple],
    dialect: Dialect,
) -> tuple[list[tuple[str, str, str]], list[tuple[tuple, int]] | None]:
    cols = [
        extract_columns(
            query.QUERY_TEXT,
            database_name=query.DATABASE_NAME.upper() if query.DATABASE_NAME else None,
            catalog_name=query.SCHEMA_NAME.upper() if query.SCHEMA_NAME else None,
            schema=info_schema,
            dialect=dialect,
        )
        for query in queries
    ]
    col_counts: Counter = summarize_columns(cols)

    # Print the most common columns in a human-readable format with one column per line
    if len(col_counts) > 0:
        max_len = 20 if len(col_counts) > 20 else len(col_counts)
        most_common_cols = col_counts.most_common(max_len)
    else:
        most_common_cols = None

    # Identify columns that are never used by comparing the columns in the info schema to the columns in the queries
    info_schema_cols = set(info_schema_flat)
    used_cols = set(col_counts.keys())
    unused_cols = sorted(info_schema_cols - used_cols)
    return unused_cols, most_common_cols


def detect_unused_tables(
    queries: list[QueryRow],
    info_schema: dict,
    info_schema_flat: list[tuple],
    dialect: Dialect,
) -> tuple[list[tuple[str, str, str]], list[tuple[tuple, int]] | None]:
    tables = [
        extract_tables(
            query.QUERY_TEXT,
            catalog_name=query.DATABASE_NAME.upper() if query.DATABASE_NAME else None,
            database_name=query.SCHEMA_NAME.upper() if query.SCHEMA_NAME else None,
            schema=info_schema,
            dialect=dialect,
        )
        for query in queries
    ]

    tbls = [item for sublist in tables for item in sublist]
    table_counts = Counter(tbls)

    # Print the most common tables in a human-readable format with one table per line
    if len(table_counts) > 0:
        max_len = 20 if len(table_counts) > 20 else len(table_counts)
        most_common_tables = table_counts.most_common(max_len)
    else:
        most_common_tables = None

    # Identify tables that are never used by comparing the tables in the info schema to the tables in the queries
    info_schema_tables = set()
    for table in info_schema_flat:
        info_schema_tables.add((table[0].upper(), table[1].upper(), table[2].upper()))

    used_tables = set(table_counts.keys())
    unused_tables = sorted(info_schema_tables - used_tables)
    return unused_tables, most_common_tables



from datetime import datetime
from enum import Enum
from typing import Any

import click
from pydantic import BaseModel


class QueryRow(BaseModel):
    QUERY_TEXT: str
    START_TIME: datetime
    DATABASE_NAME: str | None
    SCHEMA_NAME: str | None


class EnrichedQueryRow(QueryRow):
    USED_TABLES: list[str]


class SchemaRow(BaseModel):
    TABLE_CATALOG: str
    TABLE_SCHEMA: str
    TABLE_NAME: str
    COLUMN_NAME: str


class Dialect(Enum):
    snowflake = "snowflake"
    bigquery = "bigquery"
    redshift = "redshift"


def validate_dialect(ctx: Any, param: Any, value: str) -> Dialect:
    try:
        return Dialect(value)
    except ValueError:
        raise click.BadParameter(
            f'Invalid dialect value: {value}. Valid values are: {", ".join(d.value for d in Dialect)}'
        ) from None


class Grain(Enum):
    schema = "schema"
    table = "table"
    column = "column"

    def plural(self) -> str:
        return f"{self.value}s"


def validate_grain(ctx: Any, param: Any, value: str) -> Grain:
    try:
        return Grain(value)
    except ValueError:
        raise click.BadParameter(
            f'Invalid grain value: {value}. Valid values are: {", ".join(g.value for g in Grain)}'
        ) from None



import os
from datetime import date, datetime

import snowflake.connector
from pydantic import BaseModel
from snowflake.connector import SnowflakeConnection
from sqlglot import MappingSchema

from odp.core.detect_unused import extract_tables
from odp.core.types import Dialect, EnrichedQueryRow, QueryRow, SchemaRow


class SnowflakeCredentials(BaseModel):
    snowflake_account: str
    snowflake_user: str
    snowflake_password: str
    snowflake_database: str
    snowflake_warehouse: str | None = None
    snowflake_role: str | None = None


def load_snowflake_credentials() -> SnowflakeCredentials:
    """
    Load snowflake credentials from env.

    Use ODP_ prefixed vars if present, fall back to standard SNOWFLAKE_ prefix.
    """
    return SnowflakeCredentials(
        snowflake_account=os.getenv("ODP_SNOWFLAKE_ACCOUNT", os.getenv("SNOWFLAKE_ACCOUNT")),  # type: ignore[arg-type]
        snowflake_user=os.getenv("ODP_SNOWFLAKE_USERNAME", os.getenv("SNOWFLAKE_USERNAME")),  # type: ignore[arg-type]
        snowflake_password=os.getenv("ODP_SNOWFLAKE_PASSWORD", os.getenv("SNOWFLAKE_PASSWORD")),  # type: ignore[arg-type]
        snowflake_database=os.getenv("ODP_SNOWFLAKE_DATABASE", os.getenv("SNOWFLAKE_DATABASE")),  # type: ignore[arg-type]
        snowflake_warehouse=os.getenv("ODP_SNOWFLAKE_WAREHOUSE", os.getenv("SNOWFLAKE_WAREHOUSE")),
        snowflake_role=os.getenv("ODP_SNOWFLAKE_ROLE", os.getenv("SNOWFLAKE_ROLE")),
    )


def get_snowflake_connection(credentials: SnowflakeCredentials) -> SnowflakeConnection:
    return snowflake.connector.connect(
        user=credentials.snowflake_user,
        password=credentials.snowflake_password,
        account=credentials.snowflake_account,
        database=credentials.snowflake_database,
        role=credentials.snowflake_role,
        warehouse=credentials.snowflake_warehouse,
    )

def make_snowflake_mapping_schema(info_schema: dict) -> MappingSchema:
    return MappingSchema(info_schema, dialect="snowflake")


def parse_snowflake_query(query_rows: list[QueryRow], schema: MappingSchema) -> list[EnrichedQueryRow]:
    res = []
    for row in query_rows:
        used_tables = extract_tables(
            query_text=row.QUERY_TEXT,
            dialect=Dialect.snowflake,
            schema=schema,
        )
        enriched_row = EnrichedQueryRow(
            QUERY_TEXT=row.QUERY_TEXT,
            DATABASE_NAME=row.DATABASE_NAME,
            SCHEMA_NAME=row.SCHEMA_NAME,
            START_TIME=row.START_TIME,
            USED_TABLES=[".".join([part.upper() for part in table]) for table in used_tables],
        )
        res.append(enriched_row)
    return res


def get_snowflake_queries(
    conn: SnowflakeConnection,
    since_datetime: datetime | date,
    before_datetime: datetime | date,
    database_name: str | None = None,
) -> list[QueryRow]:
    # Create a cursor object.
    cur = conn.cursor()

    # Execute a statement that will generate a result set.
    sql = f"""
SELECT QUERY_TEXT, DATABASE_NAME, SCHEMA_NAME, START_TIME
FROM SNOWFLAKE.ACCOUNT_USAGE.QUERY_HISTORY
WHERE QUERY_TEXT ILIKE '%%select%%'
    {"" if database_name is None else "AND DATABASE_NAME = %(database_name)s"}
    AND START_TIME > %(since_datetime)s
    AND START_TIME < %(before_datetime)s
    AND ERROR_CODE IS NULL
ORDER BY START_TIME DESC
LIMIT 10000;
        """
    cur.execute(
        sql,
        {
            "database_name": database_name,
            "since_datetime": since_datetime,
            "before_datetime": before_datetime,
        },
    )

    return [
        QueryRow(
            QUERY_TEXT=row[0],
            DATABASE_NAME=row[1],
            SCHEMA_NAME=row[2],
            START_TIME=row[3],
        )
        for row in cur.fetchall()
    ]


def get_snowflake_schema(conn: SnowflakeConnection) -> list[SchemaRow]:
    # Create a cursor object.
    cur = conn.cursor()

    # Execute a statement that will generate a result set.
    sql = f"""
SELECT
TABLE_CATALOG,
TABLE_SCHEMA,
TABLE_NAME,
COLUMN_NAME
FROM {conn.database}.information_schema.columns
WHERE TABLE_SCHEMA != 'INFORMATION_SCHEMA';
    """
    cur.execute(sql)

    return [
        SchemaRow(
            TABLE_CATALOG=row[0],
            TABLE_SCHEMA=row[1],
            TABLE_NAME=row[2],
            COLUMN_NAME=row[3],
        )
        for row in cur.fetchall()
    ]



import os
from datetime import datetime, timedelta

import click
from dotenv import load_dotenv

from odp.core.detect_unused import (
    build_info_schema,
    detect_unused_columns,
    detect_unused_tables,
    read_info_schema_from_file,
    read_queries,
)
from odp.core.snowflake import (
    get_snowflake_connection,
    get_snowflake_queries,
    get_snowflake_schema,
    load_snowflake_credentials,
)
from odp.core.types import Dialect, Grain, validate_dialect, validate_grain

load_dotenv(dotenv_path=os.path.join(os.getcwd(), ".env"))


@click.group(name="odp")
def cli() -> None:
    pass


@cli.command(
    "detect-unused",
    help="""
Detect unused columns in SQL queries.
Requires two files: a query file and an information schema file.
Run `show-queries` to see the SQL queries to run against your
Snowflake instance to generate the two files.""",
)
@click.option("--queries-file", help="The SQL query file to analyze.")
@click.option("--schema-file", help="The file containing the information schema for the database.")
@click.option(
    "--dialect",
    type=click.Choice([d.value for d in Dialect]),
    callback=validate_dialect,
    default="snowflake",
    help="The type of warehouse to connect to. Currently only snowflake is supported.",
)
@click.option(
    "--grain",
    type=click.Choice([g.value for g in Grain]),
    callback=validate_grain,
    default="table",
    help="the grain to search for, e.g. use --grain=column to search for unused tables. Default is table.",
)
@click.option(
    "--since-days",
    type=click.INT,
    default=60,
    help="Count as 'unused' any asset that has not been queried in the last N days. Defaults to 60 days",
)
def cli_detect_unused_columns(
    queries_file: str | None,
    schema_file: str | None,
    dialect: Dialect,
    grain: Grain,
    since_days: int,
) -> None:
    if queries_file and schema_file:
        queries = read_queries(queries_file, since_days)
        print(f"Read {len(queries)} queries from {queries_file}")

        info_schema, info_schema_flat = read_info_schema_from_file(schema_file)
        print(f"Read {len(info_schema_flat)} information schema rows from {schema_file}")
        if grain == Grain.column:
            unused, most_common = detect_unused_columns(queries, info_schema, info_schema_flat, dialect)
            print_results(most_common, unused, grain)
        elif grain == Grain.table:
            unused, most_common = detect_unused_tables(queries, info_schema, info_schema_flat, dialect)
            print_results(most_common, unused, grain)
        elif grain == Grain.schema:
            raise NotImplementedError("Schema grain is not yet supported")

        return

    if dialect == Dialect.snowflake:
        try:
            credentials = load_snowflake_credentials()
        except KeyError as e:
            raise ValueError(
                f"""
Missing or invalid parameters: {e}. Please provide either

   1. both of --info-schema-file and --queries-file (use "odp show-queries" to learn how to generate these)
   2. valid credentials via env, e.g. ODP_SNOWFLAKE_ACCOUNT, ODP_SNOWFLAKE_USERNAME, etc
            """
            ) from e

        conn = get_snowflake_connection(credentials)
        schema = get_snowflake_schema(conn)
        before_date = datetime.combine(datetime.now(), datetime.max.time())
        since_date = before_date - timedelta(days=since_days)
        queries = get_snowflake_queries(
            conn,
            since_datetime=since_date,
            before_datetime=before_date,
        )

        info_schema, info_schema_flat = build_info_schema(schema)
        if grain == Grain.column:
            unused, most_common = detect_unused_columns(queries, info_schema, info_schema_flat, dialect)
            print_results(most_common, unused, grain)
        elif grain == Grain.table:
            unused, most_common = detect_unused_tables(queries, info_schema, info_schema_flat, dialect)
            print_results(most_common, unused, grain)
        elif grain == Grain.schema:
            raise NotImplementedError("Schema grain is not yet supported")
    elif dialect == Dialect.bigquery:
        raise NotImplementedError("Loading BigQuery data via credentials is not yet supported")
    elif dialect == Dialect.redshift:
        raise NotImplementedError("Loading Redshift data via credentials is not yet supported")


def print_results(
    most_common: list[tuple[tuple[str, str, str], int]] | None,
    unused: list[tuple[str, str, str]],
    grain: Grain = Grain.table,
) -> None:
    def render(obj: tuple) -> str:
        """Render a table or column tuple as a string"""
        return ".".join([x.upper() for x in obj])

    if most_common is not None:
        print(f"Most common {grain.plural()}:")
        for asset, count in most_common:
            print(f"{render(asset)}: {count}")
    else:
        print(f"No most common {grain.plural()} found in provided date range")
    print(f"Unused {grain.plural()} ({len(unused)}):")
    for asset in sorted(unused):
        print(f"{render(asset)}")


@cli.command("show-queries")
def show_snowflake_queries() -> None:
    print("Run the below against your snowflake instance to generate a dataset you can" "export to CSV for analysis")
    print(
        """
-- query_file

SELECT *
FROM SNOWFLAKE.ACCOUNT_USAGE.QUERY_HISTORY
WHERE QUERY_TEXT ILIKE 'select%'
ORDER BY START_TIME DESC
LIMIT 10000; -- or start_time > $SOME_DATE to get columns unused in the last N days


-- schema_file

use database MY_DATABASE; -- change me

SELECT
TABLE_CATALOG,
TABLE_SCHEMA,
TABLE_NAME,
COLUMN_NAME
FROM information_schema.columns
WHERE TABLE_SCHEMA != 'INFORMATION_SCHEMA';

        """
    )


if __name__ == "__main__":
    cli()
